
#include "SkeletonCalibration.hh"

#include <limits>
#include <iostream>

#include <iomanip>

#include <newmatap.h>
#include <newmatio.h>

#include <math.h>

#include "quaternion.h"

SkeletonCalibration::SkeletonCalibration(int n, int initialBufferSize)
{
	nSensors = n;
	root=0;

	p = new int[n];
	graph = new float[n*n];
	residualRMS = new float[n];
	residualMax = new float[n];
	residualAbs = new float[n];

	L = new ColumnVector[n];
	C = new ColumnVector[n];
	
	for(int i=0; i < n; i++) {
		residualRMS[i] = 0.0;
		residualMax[i] = 0.0;
		residualAbs[i] = 0.0;
	}

	joint = new ColumnVector[n];
	joint1 = new ColumnVector[n];
	joint2 = new ColumnVector[n];
	residual = new ColumnVector[n];

	ColumnVector zero(3); zero = 0.0;
	for(int i=0; i < n; i++) {
		L[i]=zero;
		C[i]=zero;

		joint[i]=zero;
		joint1[i]=zero;
		joint2[i]=zero;
		residual[i]=zero;
	}

	MM = new Matrix *[n];
	dd = new ColumnVector *[n];

	for(int i=0; i < n; i++) {
		MM[i]=0;
		dd[i]=0;
	}

	stationBuffer = new vector<vector<station_t> >;
	stationBuffer->reserve(initialBufferSize);
}


SkeletonCalibration::~SkeletonCalibration()
{
	delete stationBuffer;

	for(int i=0; i < nSensors; i++) {
		if(MM[i] != 0)
			delete MM[i];
		if(dd[i] != 0)
			delete dd[i];
	}

	delete[] dd;
	delete[] MM;

	delete[] residual;
	delete[] joint2;
	delete[] joint1;
	delete[] joint;

	delete[] C;
	delete[] L;

	delete[] residualAbs;
	delete[] residualMax;
	delete[] residualRMS;
	delete[] graph;
	delete[] p;
}

void SkeletonCalibration::setRoot(int r)
{
	root=r;
}

float dist(station_t s1, station_t s2) {
	float dx = s1.x - s2.x;
	float dy = s1.y - s2.y;
	float dz = s1.z - s2.z;
	return sqrt(dx*dx + dy*dy + dz*dz);
}

float sumDist(vector<station_t> frame1, vector<station_t> frame2)
{
	float sum=0.0;
	for(int i=0; i < frame1.size(); i++)
		sum += dist(frame1[i], frame2[i]);
	return sum;
}

void SkeletonCalibration::addData(vector<vector<station_t> > *data)
{

	// queue semantics
/*	while(data.size() > 0) {
		stationBuffer->push_back(data.front());
		data.pop();
	}*/

	float threshold = 3.0;
	
	stationBuffer->push_back((*data)[0]);
	//vector semantics
	for(unsigned int i=1; i < data->size(); i++) {
		if(sumDist(data->at(i),stationBuffer->back()) > threshold)
			stationBuffer->push_back((*data)[i]);
	}
	
	cerr << "stored " << stationBuffer->size() << " frames for calibration" << endl;
}


void SkeletonCalibration::writeCalibrationResult(ostream &stream)
{
	stream << "P";
	for(int i=0; i < nSensors; i++)
		stream << " " << p[i];
	stream << endl;
	
	stream << "R";
	for(int i=0; i < nSensors; i++)
		stream << " " << residualRMS[i];
	stream << endl;
	
	for(int i=0; i < nSensors; i++) {
		stream << "C " << i << " " << C[i].t();
		stream << "L " << i << " " << L[i].t();
	}
	stream << ";" << endl;
}


void SkeletonCalibration::readCalibrationResult(istream &stream)
{
	bool complete=false;
	int i;
	char id;
	
	ColumnVector zero(3); zero = 0.0;
	for(int i=0; i < nSensors; i++) {
		p[i] = -1;
		
		L[i]=zero;
		C[i]=zero;
	}

	while(stream && !complete) {
		stream >> id;
		switch(id) {
			case 'P':
				for(i=0; i < nSensors; i++) {
					stream >> p[i];
					if(p[i] == -1) root = i;
				}
				break;
			case 'R':
				for(i=0; i < nSensors; i++)
					stream >> residualRMS[i];
				break;
			case 'C':
				stream >> i;
				for(int j=1; j <= 3; j++)
					stream >> C[i](j);
				break;
			case 'L':
				stream >> i;
				for(int j=1; j <= 3; j++)
					stream >> L[i](j);
				break;
			case ';':
			default: // this would mean an error!
				complete=true;
				break;
		}
	}

	
}


int* SkeletonCalibration::getParenthood()
{
	return p;
}


int SkeletonCalibration::getRoot()
{
	return root;
}


float* SkeletonCalibration::getLinkageGraph()
{
	return graph;
}


vector<vertex3> SkeletonCalibration::getVertices(vector<station_t> &frame)
{
	calculateJoints(frame);

	vector<vertex3> vertices;

	// sensor positions
	for(unsigned int i=0; i < frame.size(); i++) {
		vertex3 v;
		v.v[0] = frame.at(i).x;
		v.v[1] = frame.at(i).y;
		v.v[2] = frame.at(i).z;
		
		vertices.push_back(v);
	}

	// joint positions
	for(int i=0; i < nSensors; i++) {
		vertex3 v;
		v.v[0] = joint[i](1);
		v.v[1] = joint[i](2);
		v.v[2] = joint[i](3);

		vertices.push_back(v);
	}

	return vertices;
}

vector<float> SkeletonCalibration::getRelativeErrors(vector<station_t> &frame)
{
	vector<float> errors;
	
	for(int i=0; i < nSensors; i++) {
		if(i != root)
			errors.push_back(sqrt(residual[i].NormFrobenius()) / residualRMS[i]);
	}
	
	return errors;
}


vector<int> SkeletonCalibration::getEdges()
{
	int n = nSensors;

	int deg[n];
	degree(p,deg,n);

	vector<int> edges;

	for(int i=0; i < n; i++) {
		if(i != root) {
			if(p[i] == root) {   // joint -> root
				// todo: if deg(root) > 1 then compute the center of mass between its attached joints (is better than the sensor itself)
				edges.push_back(i + nSensors);
				edges.push_back(root);
			} else {             // joint -> joint
				edges.push_back(i + nSensors);
				edges.push_back(p[i] + nSensors);
			}

			if(deg[i] == 1 && i != root) {    // joint -> leaf
				edges.push_back(i);
				edges.push_back(i + nSensors);
			}
		}
	}

	return edges;
}


vector<vertex3> SkeletonCalibration::getJoints(vector<station_t> &frame)
{
	calculateJoints(frame);

	vector<vertex3> points;
	for(int i=0; i < nSensors; i++) {
		if(i != root) {
			vertex3 v;
			v.v[0] = joint1[i](1);
			v.v[1] = joint1[i](2);
			v.v[2] = joint1[i](3);

			points.push_back(v);

			v.v[0] = joint2[i](1);
			v.v[1] = joint2[i](2);
			v.v[2] = joint2[i](3);

			points.push_back(v);
		}
	}
	return points;
}


void SkeletonCalibration::calculateJoints(vector<station_t> &frame)
{
	int n = nSensors;

	Matrix R[n];
	ColumnVector t[n];

	for(int i=0; i < n; i++) {
		station_t s = frame.at(i);
		t[i] = ColumnVector(3);
		t[i] << s.x << s.y << s.z;

		R[i] = quat2rot(s.quaternion);
	}

	ColumnVector residualAcc[n];

	bool done[n];
	bool ready[n];
	int visited[n];

	ColumnVector zero(3); zero = 0.0;
	for(int i=0; i < n; i++) {
		done[i] = false;
		ready[i] = false;
		visited[i] = 0;

		residual[i] = zero;
		residualAcc[i] = zero;
		joint[i] = zero;
		joint1[i] = zero;
		joint2[i] = zero;
	}

//	done[root] = true; // from root
	
	// start from leaves
	int deg[n];
	int nChildren[n], leftChildren[n];  // number of childs on which the node depends

	degree(p,deg,n);

	for(int i=0; i < n; i++) {
		leftChildren[i] = nChildren[i] = (i == root) ? deg[i] : deg[i]-1;
	}

	int count = 0;

	// compute joint positions from leaves to root, correcting residual errors on the way
	while(count < n-1) {
		for(int i=0; i < n; i++) {
			if(i != root) {
				//if(!done[i] && done[p[i]]) {  // from root
				if(!done[i] && leftChildren[i]==0) {
					joint1[i] = t[i] - R[i]*C[i];
					joint2[i] = t[p[i]] - R[p[i]]*L[i];
					
					joint[i] = joint1[i] - residualAcc[i];

					residualAcc[p[i]] += (1.0/nChildren[p[i]])*(joint2[i] - joint[i]);
					residual[i] = joint2[i] - joint1[i];
					
					//joint2[i] -= residual[i];

					done[i]=true;     // this node was calculated
					leftChildren[p[i]]--;  // p[i] depends on one child less

					count++;
				}
			}
		}
	}
}


void SkeletonCalibration::makeCalibrationMatrix(int n, Matrix **M, ColumnVector **d, int strive)
{
	int size = stationBuffer->size()/strive;

	Matrix *M_ = new Matrix(size*3,3);
	ColumnVector *d_ = new ColumnVector(size*3);

	for(int i=0; i < size/strive; i++) {
		station_t p = stationBuffer->at(i*strive).at(n);

		Matrix R = quat2rot(p.quaternion);
		M_->SubMatrix(3*i+1, 3*i+3, 1, 3) = R;

		(*d_)(i*3+1 + 0) = p.x;
		(*d_)(i*3+1 + 1) = p.y;
		(*d_)(i*3+1 + 2) = p.z;
	}
	*M = M_;
	*d = d_;
}


ColumnVector SkeletonCalibration::solveCalibration(Matrix &Q1, Matrix &Q2, ColumnVector &d1, ColumnVector &d2)
{
	Matrix Q = Q1 | -Q2;
	ColumnVector d = d1 - d2;

	// now solve M*x = d

	Matrix U(Q.Nrows(), 6), V(6,6);
	DiagonalMatrix D(6);
	ColumnVector res(6);

	SVD(Q, D, U, V);
	ColumnVector dp = U.t()*d;

	for(int j=1; j <= 6; j++) {
		dp(j) = dp(j) / D(j);
	}

	return V * dp;
}

int SkeletonCalibration::guessRoot()
{
	const float fInf = std::numeric_limits<float>::max();

	const int n = nSensors;
	double power[n];
	float minPower=fInf;
	int min=0;

	int size = stationBuffer->size();

	cerr << "power";

	for(int i=0; i < n; i++) {
		power[i] = 0.0;

		station_t ps = stationBuffer->at(0).at(i);
		double x = ps.x, y = ps.y, z = ps.z;
		
		for(int j=1; j < size; j++) {
			station_t p = stationBuffer->at(j).at(i);
			power[i] += sqrt((p.x-x)*(p.x-x) + (p.y-y)*(p.y-y) + (p.z-z)*(p.z-z));
			x = p.x; y = p.y; z = p.z;
		}

		if(minPower > power[i]) {
			minPower = power[i];
			min = i;
		}
		cerr << " " << power[i];
	}
	cerr << endl;
	root = min;
	return min;
}

void SkeletonCalibration::calibrateSkeleton()
{
	int n = nSensors;  // number of sensors

	for(int i=0; i < n; i++) {
		if(MM[i] != 0) delete MM[i];
		if(dd[i] != 0) delete dd[i];
	}

	for(int i=0; i < n; i++) 
		makeCalibrationMatrix(i, &MM[i], &dd[i], 1);  // store motion data in matrix arrays MM and dd

	float (*G)[n] = (float (*)[n]) graph;
	for(int i=0; i < n; i++)
		for(int j=0; j < n; j++)
			G[i][j] = 0;
	
	float resMax[n][n];
	float resAbs[n][n];
	
	for(int i=0; i < n; i++) {
		for(int j=0; j < n; j++) {
			resMax[i][j] = 0.0;
			resAbs[i][j] = 0.0;
		}
	}
	

	ColumnVector results[n][n];  // nxn array of column vectors representing the calibration results
	for(int i=0; i < n; i++) {
		for(int j=0; j < i; j++) {
			ColumnVector t = solveCalibration(*MM[j], *MM[i], *dd[j], *dd[i]);
			ColumnVector residual = (*MM[j] | -(*MM[i]))*t - (*dd[j] - *dd[i]);

			double max=0.0;
			double sum=0.0;
			for(int k=1; k < residual.Nrows()-2; k+=3) {
				double x = residual(k+0);
				double y = residual(k+1);
				double z = residual(k+2);
				double dist = sqrt(x*x + y*y + z*z);
				max = (dist > max) ? dist : max;
				sum += dist;
			}
			resMax[j][i] = max;
			resAbs[j][i] = sum/residual.Nrows()*3;

			G[j][i] = sqrt(3.0/MM[i]->Nrows())*residual.NormFrobenius();  // calculate RMS error

			results[i][j] = t;
		}
	}
	
	// store the RESULT of the calibration (quite fast!)

	// find the minimum spanning tree (our skeleton)
	prim((float*) G, p, n, root);

	// store results
	for(int i=0; i < n; i++) {
		if(i != root) {
			L[i] = (p[i] < i) ? results[i][p[i]].SubMatrix(1,3,1,1)
			                  : results[p[i]][i].SubMatrix(4,6,1,1);

			C[i] = (p[i] < i) ? results[i][p[i]].SubMatrix(4,6,1,1)
			                  : results[p[i]][i].SubMatrix(1,3,1,1);

			residualRMS[i] = (p[i] < i) ? G[p[i]][i] : G[i][p[i]];
			residualMax[i] = (p[i] < i) ? resMax[p[i]][i] : resMax[i][p[i]];
			residualAbs[i] = (p[i] < i) ? resAbs[p[i]][i] : resAbs[i][p[i]];
		}
	}


	// debug output
	cerr << "p =";
	for(int i=0; i < n; i++) {
		cout << " " << p[i];
	} cout << endl;

	cerr << "G =" << endl;
	for(int i=0; i < n; i++) {
		for(int j=0; j < n; j++) {
			cerr << " " << setw(9) << G[i][j];
		} cerr << endl;
	}

	cerr << "residualRMS =";
	for(int i=0; i < n; i++) {
		cout << " " << residualRMS[i];
	} cout << endl;
	
	cerr << "residualMax =";
	for(int i=0; i < n; i++) {
		cout << " " << residualMax[i];
	} cout << endl;
	
	cerr << "residualAbs =";
	for(int i=0; i < n; i++) {
		cout << " " << residualAbs[i];
	} cout << endl;
}


// utiliy functions


void SkeletonCalibration::prim(float *G, int *p, int n, int root)
{
	const float fInf = std::numeric_limits<float>::max();

	float (*GG)[n] = (float (*)[n]) G;

	// helper vectors
	int source[n], destination[n], listV[n];
	for(int i=0; i < n; i++) {
		source[i]=destination[i]=p[i]=-1;
		listV[i]=0;
	}
	listV[root]=1;

	//g=G+G'
	float g[n][n];
	for(int i=0; i < n; i++)
		for(int j=0; j < n; j++)
			g[i][j] = (GG[i][j]==0 && GG[j][i]==0) ? fInf : GG[i][j] + GG[j][i];

	for(int e=0; e<n-1; e++) {
		float min=fInf;
		int s=-1, d=-1;
		for(int i=0; i < n; i++)
			if(listV[i]==1)
				for(int j=0; j < n; j++)
					if(listV[j]==0)
						if(min>g[i][j]) {
							min=g[i][j];
							s=i;
							d=j;
						}

		listV[d]=1;
		source[e]=s;
		destination[e]=d;
	}

	for(int i=0; i < n; i++)
		if(destination[i] != -1)
			p[destination[i]] = source[i];
}


void SkeletonCalibration::degree(int *tree, int *deg, int n)
{
	for(int i=0; i < n; i++) {
		deg[i]=0;
	}

	for(int i=0; i < n; i++) {
		if(tree[i] >= 0) {
			deg[i]++;        // outbound edge
			deg[tree[i]]++;  // inbound edge
		}
	}
}

/*
inline Matrix SkeletonCalibration::quat2rot(float *q)
{
	Matrix rot(3,3);

	rot(1,1) = q[0]*q[0] + q[1]*q[1] - q[2]*q[2] - q[3]*q[3];
	rot(1,2) = 2*(q[1]*q[2] - q[0]*q[3]);
	rot(1,3) = 2*(q[1]*q[3] + q[0]*q[2]);

	rot(2,1) = 2*(q[1]*q[2] + q[0]*q[3]);
	rot(2,2) = q[0]*q[0] - q[1]*q[1] + q[2]*q[2] - q[3]*q[3];
	rot(2,3) = 2*(q[2]*q[3] - q[0]*q[1]);

	rot(3,1) = 2*(q[1]*q[3] - q[0]*q[2]);
	rot(3,2) = 2*(q[2]*q[3] + q[0]*q[1]);
	rot(3,3) = q[0]*q[0] - q[1]*q[1] - q[2]*q[2] + q[3]*q[3];

	return rot;
}

float* rot2quat(const Matrix a,float *q)
{
	const float fEps = std::numeric_limits<float>::epsilon();
	
	float trace = a(1,1) + a(2,2) + a(3,3) + 1.0f;
	if( trace > fEps ) {
		float s = 0.5f / sqrtf(trace);
		q[0] = 0.25f / s;
		q[1] = ( a(3,2) - a(2,3) ) * s;
		q[2] = ( a(1,3) - a(3,1) ) * s;
		q[3] = ( a(2,1) - a(1,2) ) * s;
	} else {
		if ( a(1,1) > a(2,2) && a(1,1) > a(3,3) ) {
			float s = 2.0f * sqrtf( 1.0f + a(1,1) - a(2,2) - a(3,3));
			q[0] = (a(2,3) - a(3,2) ) / s;
			q[1] = 0.25f * s;
			q[2] = (a(1,2) + a(2,1) ) / s;
			q[3] = (a(1,3) + a(3,1) ) / s;
		} else if (a(2,2) > a(3,3)) {
			float s = 2.0f * sqrtf( 1.0f + a(2,2) - a(1,1) - a(3,3));
			q[0] = (a(1,3) - a(3,1) ) / s;
			q[1] = (a(1,2) + a(2,1) ) / s;
			q[2] = 0.25f * s;
			q[3] = (a(2,3) + a(3,2) ) / s;
		} else {
			float s = 2.0f * sqrtf( 1.0f + a(3,3) - a(1,1) - a(2,2) );
			q[0] = (a(1,2) - a(2,1) ) / s;
			q[1] = (a(1,3) + a(3,1) ) / s;
			q[2] = (a(2,3) + a(3,2) ) / s;
			q[3] = 0.25f * s;
		}
	}
	return q;
}
*/
