
#include "SkeletonViewer.hh"
#include "SkeletonCalibration.hh"

#include <math.h>
#include <GL/gl.h>
#include <GL/glu.h>

#ifndef PI
#define PI 3.14159265358979323846
#endif


SkeletonViewer::SkeletonViewer(SkeletonCalibration *skel)
{
	skeleton = skel;
	_quad = gluNewQuadric();

	errorRange = 3.0;
}


SkeletonViewer::~SkeletonViewer()
{
	gluDeleteQuadric(_quad);
}


void SkeletonViewer::setFrame(vector<station_t> frame)
{
	currentFrame = frame;
}


// set up lighting and material
void SkeletonViewer::setupLighting() {
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	
	GLfloat light_pos[] = {35.0, -35.0, -35.0, 1.0};
	GLfloat light_Ka[] = {0.0, 0.0, 0.0, 1.0};
	GLfloat light_Kd[] = {1.0, 1.0, 1.0, 1.0};
	GLfloat light_Ks[] = {1.0, 1.0, 1.0, 1.0};
		
	glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
	glLightfv(GL_LIGHT0, GL_AMBIENT, light_Ka);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_Kd);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light_Ks);
	
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE); //glColor changes the diffuse color
	
	GLfloat material_Ka[] = {0.12, 0.12, 0.12, 1.0};
	GLfloat material_Kd[] = {0.5, 0.5, 0.5, 1.0};
	GLfloat material_Ks[] = {0.3, 0.3, 0.3, 1.0};
	GLfloat material_Ke[] = {0.0, 0.0, 0.0, 1.0};
	GLfloat material_Se = 10;
	
	glMaterialfv(GL_FRONT, GL_AMBIENT, material_Ka);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, material_Kd);
	glMaterialfv(GL_FRONT, GL_SPECULAR, material_Ks);
	glMaterialfv(GL_FRONT, GL_EMISSION, material_Ke);
	glMaterialf(GL_FRONT, GL_SHININESS, material_Se);
}


void SkeletonViewer::glInit()
{
	setupLighting();
	
	glClearDepth(1.0);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
}


void SkeletonViewer::drawCoordinateSystem(float size)
{
	glDisable(GL_LIGHTING);
	glBegin(GL_LINES);
	
	glColor3f(1, 0, 0);
	glVertex3f(0, 0, 0);
	glVertex3f(size, 0, 0);
	
	glColor3f(0, 1, 0);
	glVertex3f(0, 0, 0);
	glVertex3f(0, size, 0);
	
	glColor3f(0, 0, 1);
	glVertex3f(0, 0, 0);
	glVertex3f(0, 0, size);
	
	glEnd();
	glEnable(GL_LIGHTING);
}


void SkeletonViewer::drawCylinder(float *from, float *to, float size)
{
	float dir[3];

	dir[0] = to[0] - from[0];
	dir[1] = to[1] - from[1];
	dir[2] = to[2] - from[2];

	float l2 = dir[0]*dir[0] + dir[1]*dir[1] + dir[2]*dir[2];
	float len = sqrt(l2);
	float s = sqrt(l2/(l2 - dir[0]*dir[0]));

	dir[0] *= 1.0/len;
	dir[1] *= 1.0/len;
	dir[2] *= 1.0/len;

	GLfloat matrix[16];
	matrix[0] = s*(dir[0]*dir[0] - 1.0); matrix[1] = s*dir[0]*dir[1]; matrix[2] = s*dir[0]*dir[2];
	matrix[3] = 0.0;
	matrix[4] = 0.0; matrix[5] = -dir[2]*s; matrix[6] = dir[1]*s;
	matrix[7] = 0.0;
	matrix[8] = dir[0]; matrix[9] = dir[1]; matrix[10] = dir[2];
	matrix[11] = 0.0;
	matrix[12] = from[0]; matrix[13] = from[1]; matrix[14] = from[2]; matrix[15] = 1.0;

	glPushMatrix();

	glMultMatrixf(matrix);

	// attribute in new class!!!
//	GLUquadric* quad = gluNewQuadric();  // create helper object

	gluCylinder(_quad, size, size, len, 10, 2);

	glPopMatrix();
}

void SkeletonViewer::drawCube(float size) {
	float vertex[][3] = {{1,1,-1},{1,1,1},{-1,1,1},{-1,1,-1}  ,{1,-1,-1},{1,-1,1},{-1,-1,1},{-1,-1,-1}};
	int face[][4] = {{0,3,2,1},{0,4,7,3},{0,1,5,4},{2,3,7,6},{1,2,6,5},{4,5,6,7}};
	float normal[][3] = {{0,1,0},{0,0,-1},{1,0,0},{-1,0,0},{0,0,1},{0,-1,0}};

	for(int v=0; v < 8; v++)
		for(int i=0; i < 3; i++)
			vertex[v][i] *= size*0.5;

	glBegin(GL_QUADS);
	for(int f=0; f < 6; f++) {
		glNormal3fv(normal[f]);		
		for(int i=0; i < 4; i++) {
			glVertex3fv(vertex[face[f][i]]);
		}
	}
	glEnd();
}

//float _color[3];

float *SkeletonViewer::errorColor(float e)
{
	_color[0] = e;
	_color[1] = 1.0 - e;
	_color[2] = 0.0;

	return _color;
}

float *SkeletonViewer::errorColor2(float e)
{
	_color[0] = e;
	_color[1] = 0.0;
	_color[2] = 1.0 - e;

	return _color;
}

void SkeletonViewer::glDraw()
{
	if(currentFrame.size() == 0)
		return;

//	GLUquadric* _quad = gluNewQuadric();  // create helper object	
	drawCoordinateSystem(8);
	glColor3f(0.3, 0.3, 0.3);
	drawCube(5);

	vector<vertex3> joints = skeleton->getJoints(currentFrame);

	vector<float> error = skeleton->getRelativeErrors(currentFrame);

/*	cerr << "E:";
	for(unsigned int i=0; i < error.size(); i++) {
		cerr << " " << error[i];
	} cerr << endl;
*/
	glColor3f(0.0, 1.0, 0.0);
	for(unsigned int i=0; i < currentFrame.size(); i++) {
		glPushMatrix();
		glTranslatef(currentFrame[i].x, currentFrame[i].y, currentFrame[i].z);
		float *q = currentFrame[i].quaternion;
		float scale = sqrt(q[1]*q[1] + q[2]*q[2] + q[3]*q[3]);
		glRotatef(acos(q[0])*2.0/PI*180, q[1]/scale, q[2]/scale, q[3]/scale);
		glColor3f(0.0, 1.0, 0.0);
		drawCube(3.0);
		drawCoordinateSystem(5);
		glPopMatrix();
	}

	for(unsigned int i=0; i < joints.size()/2; i++) {
		//glColor3f(0.0, 1.0, 0.0);
		glColor3fv(errorColor(error[i] / errorRange));
		glPushMatrix();
		glTranslatef(joints.at(i*2).v[0], joints.at(i*2).v[1], joints.at(i*2).v[2]);
		gluSphere(_quad,3.0,10,10);
		glPopMatrix();

		glColor3f(1.0, 1.0, 0.0);
		glColor3fv(errorColor(error[i] / errorRange));
		glPushMatrix();
		glTranslatef(joints.at(i*2+1).v[0], joints.at(i*2+1).v[1], joints.at(i*2+1).v[2]);
		gluSphere(_quad,3.0,10,10);
		glPopMatrix();
	}

	vector<vertex3> vertex = skeleton->getVertices(currentFrame);
	vector<int> edge = skeleton->getEdges();


	glColor3f(1.0, 1.0, 1.0);
	for(unsigned int i=0; i < edge.size()/2; i++) {
		//glBegin(GL_LINES);
		//glVertex3f(vertex[edge[i*2+0]].v[0], vertex[edge[i*2+0]].v[1], vertex[edge[i*2+0]].v[2]);
		//glVertex3f(vertex[edge[i*2+1]].v[0], vertex[edge[i*2+1]].v[1], vertex[edge[i*2+1]].v[2]);
		//glEnd();

		drawCylinder(vertex[edge[i*2+0]].v, vertex[edge[i*2+1]].v, 1.0);
	}
}
